;
; LIBRARY.ASM  -  MagicKit Standard Library
;
;

; IMPORTANT NOTE:
; ----
; almost all the library functions have been changed to automaticaly
; handle bank mapping (you don't have to map data yourself anymore),
; the change will be transparent to you if you were using only library
; macros to call functions, but you will have to update your code
; in case you were directly calling those functions!


; ----
; map_data
; ----
; map data in page 3-4 ($6000-$9FFF)
; ----
; IN :  _BL = data bank
;       _SI = data address
; ----
; OUT:  _BX = old banks
;       _SI = remapped data address
; ----

map_data:
	ldx   <_bl

    ; ----
    ; save current bank mapping
    ;
	tma   #3
	sta   <_bl
	tma   #4
	sta   <_bh

    ; ----
    ; map new bank
    ;
	txa
	tam   #3
	inc   A
	tam   #4

    ; ----
    ; remap data address to page 3
    ;
	lda   <_si+1
	and   #$1F
	ora   #$60
	sta   <_si+1
	rts


; ----
; unmap_data
; ----
; IN :  _BX = old banks
; ----

unmap_data:

	lda	<_bl
	tam	#3
	lda	<_bh
	tam	#4
	rts

; ----
; remap_data
; ----

remap_data:
	lda   <_si+1
	bpl   .l1
	sub   #$20
	sta   <_si+1
	tma   #4
	tam   #3
	inc   A
	tam   #4
.l1:
	rts


; ----
; load_palette
; ----
; initialize one or more sub-palette
; ----
; IN :  _AL = index of the first sub-palette (0-31)
;       _BL = data bank
;       _SI = address of data
;       _CL = number of sub-palette to copy
; ----

load_palette:

    ; ----
    ; map data
    ;
	jsr   map_data

    ; ----
    ; multiply the sub-palette index by 16
    ; and set the VCE color index register
    ;
	lda   <_al
	stz   <_ah
	asl   A
	asl   A
	asl   A
	asl   A
	rol   <_ah
	sta   color_reg_l
	lda   <_ah
	sta   color_reg_h

    ; ----
    ; load new colors
    ;
	cly
.loop_a:
	ldx   #16
.loop_b:
	lda   [_si],Y		; set color
	iny
	sta   color_data_l
	lda   [_si],Y
	sta   color_data_h
	iny
	bne   .next
	inc   <_si+1
.next:  dex			; next color
	bne   .loop_b

	dec   <_cl		; next sub-palette
	bne   .loop_a

    ; ----
    ; unmap data
    ;
	jmp   unmap_data


; ----
; load_bat
; ----
; transfer a BAT in VRAM
; ----
; IN :  _DI = VRAM base address
;       _DX = width of the virtual screen
;       _BL = BAT bank
;       _SI = BAT memory location
;       _CL = nb of column to copy
;       _CH = nb of row
; ----

load_bat:

    ; ----
    ; map data
    ;
	jsr   map_data

    ; ----
    ; copy BAT
    ;
	cly
	; --
.l1:	jsr   set_write
	ldx   <_cl
	; --
.l2:	lda   [_si],Y
	sta   video_data_l
	iny
	lda   [_si],Y
	sta   video_data_h
	iny
	bne   .l3
	inc   <_si+1
.l3:	dex
	bne   .l2
	; --
	jsr   remap_data
	; --
	addw  <_dx,<_di
	dec   <_ch
	bne   .l1

    ; ----
    ; unmap data
    ;
	jmp   unmap_data


; ----
; load_map
; ----
; transfer a tiled map in VRAM
; ----
; IN :  _DI = VRAM base address
;       _CL = x start coordinate in the map
;       _CH = y start coordinate
;       _DL = nb of column to copy
;       _DH = nb of row
; ----

	.bss
mapbase       .ds 2
mapbank       .ds 1
mapwidth      .ds 1
mapheight     .ds 1
maptilebase   .ds 2
mapctable     .ds 2
mapctablebank .ds 1
	.code

load_map:

    ; ----
    ; save bank mapping
    ;
	tma   #2
	pha
	tma   #3
	pha
	tma   #4
	pha
	tma   #5
	pha

    ; ----
    ; calculate map address
    ;
	stw   mapctable,<_bp
	stw   mapbase,<_si
	and   #$1F
	sta   <_si+1
	; --
	ldx   <_cl
	ldy   <_ch
	stx   <_ah
	sty   <_bh
	sty   <_al
	lda   mapwidth
	sta   <_bl
	jsr   mulu8
	addw  <_cx,<_si

    ; ----
    ; calculate map bank
    ;
    	bcc   .l1
	lda   #8
	bra   .l2
	; --
.l1:	lda   <_si+1
	and   #$E0
	beq   .l2
	lsr   A
	lsr   A
	lsr   A
	lsr   A
	lsr   A
	; --
.l2:	add   mapbank
	sta   <_bl

    ; ----
    ; map data
    ;
    	tam   #2
	inc   A
	tam   #3
	lda   mapctablebank
	tam   #4
	inc   A
	tam   #5

    ; ----
    ; adjust data addresses
    ;
	lda   <_si+1		; map data
	and   #$1F
	ora   #$40
	sta   <_si+1
	; --
	lda   mapctable+1	; tile color table
	and   #$1F
	ora   #$80
	sta   mapctable+1

    ; ----
    ; vertical loop
    ;
.l3:	ldy   <_ah
	lda   <_dl
	sta   <_al

    ; ----
    ; horizontal loop
    ;
.l4:	lda   [_si],Y		; get tile index
	iny
	; --
	cpy   mapwidth		; horizontal wrapping
	blo   .l5
	cly

.l5:	tax			; calculate BAT value (tile + palette)
	sxy
        stz   <_ch
	asl   A
	rol   <_ch
	asl   A
	rol   <_ch
	add   maptilebase
	sta   <_cl
	lda   <_ch
	adc   maptilebase+1
	adc   [_bp],Y
	sta   <_ch
	sxy

	vreg  #0		; copy tile
	stw   <_di,video_data
	vreg  #2
	stw   <_cx,video_data
	incw  <_cx
	stw   <_cx,video_data
	incw  <_cx
	vreg  #0
	lda   batwidth
	add   <_di
	sta   video_data_l
	cla
	adc   <_di+1
	sta   video_data_h
	vreg  #2
	stw   <_cx,video_data
	incw  <_cx
	stw   <_cx,video_data

	incw  <_di		; next tile
	incw  <_di
	dec   <_al
	lbne  .l4

    ; ----
    ; next line
    ;
	lda   batwidth		; incremente vram address
	sub   <_dl
	asl   A
	add   <_di
	sta   <_di
	cla
	adc   <_di+1
	sta   <_di+1

	inc   <_bh		; vertical wrapping
	lda   <_bh
	cmp   mapheight
	blo   .l6
	; --
	lda   mapbank
	tam   #2
	inc   A
	tam   #3
	stb   mapbase,<_si
	lda   mapbase+1
	and   #$1F
	ora   #$40
	sta   <_si+1
	stz   <_bh
	bra   .l7

.l6:	lda   <_si		; incremente map address
	add   mapwidth
	sta   <_si
	cla
	adc   <_si+1
	sta   <_si+1
	; --
	cmp   #$60		; adjust bank mapping if necessary
	blo   .l7
	sub   #$20
	sta   <_si+1
	tma   #3
	tam   #2
	inc   A
	tam   #3

.l7:	dec   <_dh
	lbne  .l3

    ; ----
    ; restore bank mapping
    ;
    	pla
	tam   #5
	pla
	tam   #4
	pla
	tam   #3
	pla
	tam   #2
	rts


; ----
; load_font
; ----
; transfer a 8x8 monochrome font into VRAM, slow but can be useful
; ----
; IN :  _DI = VRAM base address
;       _BL = font bank
;       _SI = font memory location
;       _AL = font color (0-15)
;       _CL = number of characters to copy
; ----

load_font:

    ; ----
    ; map data
    ;
	jsr   map_data

    ; ----
    ; character loop
    ;
	jsr   set_write
.copy:

    ; ----
    ; plane 1
    ;
	cly
.l1:	lda   [_si],Y
	tst   #$01,<_al
	bne   .p1
	cla
.p1:	sta   video_data_l

    ; ----
    ; plane 2
    ;
	lda   [_si],Y
	tst   #$02,<_al
	bne   .p2
	cla
.p2:	sta   video_data_h

	iny
	cpy   #8
	bne   .l1

    ; ----
    ; plane 3
    ;
	cly
.l2:	lda   [_si],Y
	tst   #$04,<_al
	bne   .p3
	cla
.p3:	sta   video_data_l

    ; ----
    ; plane 4
    ;
	lda   [_si],Y
	tst   #$08,<_al
	bne   .p4
	cla
.p4:	sta   video_data_h

	iny
	cpy   #8
	bne   .l2

    ; ----
    ; next character
    ;
	addw  #8,<_si
	; --
	dec   <_cl
	bne   .copy

    ; ----
    ; unmap data
    ;
	jmp   unmap_data


; ----
; load_vram
; ----
; copy a block of memory to the VRAM
; ----
; IN :  _DI = VRAM location
;       _BL = data bank
;       _SI = data memory location
;       _CX = number of words to copy
; ----

load_vram:

    ; ----
    ; map data
    ;
	jsr   map_data

    ; ----
    ; set vram address
    ;
	jsr   set_write

    ; ----
    ; copy data
    ;
	cly
	ldx   <_cl
	beq   .l3
	; --
.l1:	lda   [_si],Y
	sta   video_data_l
	iny
	lda   [_si],Y
	sta   video_data_h
	iny
	bne   .l2
	inc   <_si+1
	; --
.l2:	dex
	bne   .l1
	; --
	jsr   remap_data
	; --
.l3:	dec   <_ch
	bpl   .l1

    ; ----
    ; unmap data
    ;
	jmp   unmap_data


; ----
; set_read
; ----
; set the VDC VRAM read pointer
; ----
; IN :  _DI = VRAM location
; ----

set_read:
	vreg  #$01
	lda   <_di 
	sta   video_data_l
	lda   <_di+1
	sta   video_data_h
	vreg  #$02
	rts 


; ----
; set_write
; ----
; set the VDC VRAM write pointer
; ----
; IN :  _DI = VRAM location
; ----

set_write:
	vreg  #$00
	lda   <_di 
	sta   video_data_l
	lda   <_di+1
	sta   video_data_h
	vreg  #$02
	rts 


	.if (!CDROM)
; ----
; HSR(xres)
; ----
; macros to calculate the value of the HSR VDC register
; ----
; IN :  xres, horizontal screen resolution
; ----

HSR	.macro
	 .if (\1 <= 272)
	  ; low res
	  .db $02
	  .db (18 - (((\1 / 8) - 1) / 2))
	 .else
	  ; high res
	  .db $03
	  .db (24 - (((\1 / 8) - 1) / 2))
	 .endif
	.endm


; ----
; HDR(xres)
; ----
; macros to calculate the value of the HDR VDC register
; ----
; IN :  xres, horizontal screen resolution
; ----

HDR	.macro
	 .db ((\1 / 8) - 1)
	 .if (\1 <= 272)
	  ; low res
	  .db (38 - ((18 - (((\1 / 8) - 1) / 2)) + (\1 / 8)))
	 .else
	  ; high res
	  .db (54 - ((24 - (((\1 / 8) - 1) / 2)) + (\1 / 8)))
	 .endif
	.endm


; ----
; init_vdc
; ----
; initialize the video controller
;  - 256x240 screen mode
;  - 64x32 virtual bgmap size
;  - display and sprites off
;  - interrupts disabled
;  - SATB at $7F00
;  - VRAM cleared
; ----

	.bss
batwidth  .ds 1
batheight .ds 1
	.code

init_vdc:

    ; ----
    ; default screen resolution
    ;
	.ifndef _xres
_xres    .equ 256
	.endif

    ; ----
    ; initialize the VDC registers
    ;
	stw   #.table,<_si 	; register table address in '_si'
	cly 
.l1:	lda   [_si],Y		; select the VDC register
	iny
	sta   <vdc_reg
	sta   video_reg
	lda   [_si],Y		; send the 16-bit data
	iny 
	sta   video_data
	lda   [_si],Y
	iny 
	sta   video_data+1
	cpy   #36		; loop if not at the end of the
	bne   .l1		; table

    ; ----
    ; set the pixel clock frequency
    ;
	.if (_xres <= 272)
	 stz  color_ctrl
	.else
	 lda  #1
	 sta  color_ctrl
	.endif

    ; ----
    ; set the background & border colors to black
    ;
	stw   #256,color_reg
	stwz  color_data
	stwz  color_reg
	stwz  color_data

    ; ----
    ; clear the video RAM
    ;
	st0   #0
	st1   #0
	st2   #0
	st0   #2

	ldx   #128
.l2:	cly
.l3:	st1   #0
	st2   #0
	dey
	bne   .l3
	dex
	bne   .l2

    ; ----
    ; set BAT size variables
    ;
	lda   #64
	sta   batwidth
	lda   #32
	sta   batheight

	rts

    ; ----
    ; VDC register table
    ;
.table:	.db $05,$00,$00		; CR    control register
	.db $06,$00,$00		; RCR   scanline interrupt counter
	.db $07,$00,$00		; BXR   background horizontal scroll offset
	.db $08,$00,$00		; BYR        "     vertical     "      "
	.db $09,$10,$00		; MWR   size of the virtual screen
	.db $0A			; HSR +
	 HSR _xres		;     |                 [$02,$02]
	.db $0B			; HDR | display size
	 HDR _xres		;     | and synchro     [$1F,$04]
	.db $0C,$02,$0F		; VPR |
	.db $0D,$EF,$00		; VDW |
	.db $0E,$04,$00		; VCR +
	.db $0F,$10,$00		; DCR   DMA control register
	.db $13,$00,$7F		; SATB  address of the SATB
	.endif


; ----
; init_psg
; ----
; initialize the sound generator.
; ----

init_psg:
	stz   psg_mainvol	; main volume to zero
	stz   psg_lfoctrl	; disable the LFO
	
	lda   #5		; set volume to zero for each channel
.clear:	sta   psg_ch            ; and disable them
	stz   psg_ctrl
	stz   psg_pan
	dec   A
	bpl   .clear

	lda   #4		; disable noise for channel 5 & 6
	sta   psg_ch
	stz   psg_noise
	lda   #5
	sta   psg_ch
	stz   psg_noise
	rts

